/* 
 * File:   XrdDPMCommon.hh
 * Author: Fabrizio Furano (CERN IT/GT), David Smith
 * 
 * Common functions
 *
 * Created on February 2, 2012, 1:57 PM
 */

#ifndef XRDDPMCOMMON_HH
#define	XRDDPMCOMMON_HH

#include <pthread.h>
#include <string.h>
#include <time.h>
#include <vector>
#include <list>
#include <memory>

#include <XrdSec/XrdSecEntity.hh>
#include <XrdOuc/XrdOucString.hh>
#include <XrdOuc/XrdOucEnv.hh>
#include <XrdOuc/XrdOucName2Name.hh>
#include <XrdSys/XrdSysError.hh>
#include <XrdSys/XrdSysPthread.hh>

#include <dmlite/cpp/dmlite.h>
#include <dmlite/cpp/utils/extensible.h>
#include <dmlite/cpp/poolmanager.h>

#define XRDDPM_IS_PUT 1
#define XRDDPM_ID_URL 2

/**********************************/
/*           classes              */
/**********************************/

class DpmIdentityConfigOptions {
public:
	XrdOucString	principal;
	std::vector<XrdOucString>	fqans;
	std::vector<XrdOucString>	validvo;
};


class DpmIdentity {
public:
	DpmIdentity(XrdOucEnv *Env, DpmIdentityConfigOptions &config);
	DpmIdentity(XrdOucEnv *Env);
	DpmIdentity();
	void	CopyToStack(dmlite::StackInstance&) const;
	const XrdOucString&	Dn() const {return m_name;}
	const XrdOucString&	Groups() const {return m_endors_raw;}
	bool	IsSecEntID() const {return UsesSecEntForID;}
	static bool	usesPresetID(XrdOucEnv *Env,
		const XrdSecEntity *Entity=0);
	static bool	badPresetID(DpmIdentityConfigOptions &config,
                XrdOucString &err);

private:
	void	parse_grps();
	void	parse_secent(const XrdSecEntity *secEntity);
	void	check_validvo(DpmIdentityConfigOptions &config);
	XrdOucString	m_name;
	std::vector<XrdOucString>	m_vorgs;
	std::vector<XrdOucString>	vs;
	XrdOucString	m_endors_raw;
	bool	UsesSecEntForID;
};

class XrdDmStackStore {
public:
	XrdDmStackStore() { }
	~XrdDmStackStore();
	static void	unsetStackParams(dmlite::StackInstance &);
	std::auto_ptr<dmlite::StackInstance>	getStack(DpmIdentity&);
	void	RetireStack(std::auto_ptr<dmlite::StackInstance>);
	void SetDmConfFile(XrdOucString &fn) { DmConfFile = fn; }

private:
	XrdDmStackStore(const XrdDmStackStore&);
	XrdDmStackStore&	operator=(const XrdDmStackStore&);
	std::auto_ptr<dmlite::PluginManager>	managerP;
	XrdSysMutex	ManagerMtx;
	XrdOucString	DmConfFile;
};

class XrdDmStackWrap {
public:
	XrdDmStackWrap(XrdDmStackStore &stks, DpmIdentity &ident) : sf(&stks) {
		siP = sf->getStack(ident);
	}

	XrdDmStackWrap() : sf(0) { }

	void reset() {
		if (siP.get()) { sf->RetireStack(siP); }
		sf = 0;
	}

	void reset(XrdDmStackStore &stks, DpmIdentity &ident) {
		if (siP.get()) { sf->RetireStack(siP); }
		sf = &stks;
		siP = sf->getStack(ident);
	}

	dmlite::StackInstance*	operator->() {
		if (!siP.get())
			throw dmlite::DmException(DMLITE_SYSERR(EINVAL),
				"No stack");
		return siP.get();
	}

	dmlite::StackInstance&	operator*() {
		if (!siP.get())
			throw dmlite::DmException(DMLITE_SYSERR(EINVAL),
				"No stack");
		return *siP;
	}

	~XrdDmStackWrap() {
		try {
			if (siP.get()) { sf->RetireStack(siP); }
		} catch(...) {
			// ignore
		}
	}

private:
	XrdDmStackWrap(const XrdDmStackWrap&);
	XrdDmStackWrap&	operator=(const XrdDmStackWrap&);
	XrdDmStackStore *sf;
	std::auto_ptr<dmlite::StackInstance> siP;
};


/**********************************/
/*          functions             */
/**********************************/

XrdOucString DecodeString(XrdOucString in);

// Encode an array of bytes to base64
char *Tobase64(const unsigned char *input, int length);


// Calculates the url hash, v1 and v2, from the given information
void calc2Hashes(
	char **hashes,
	unsigned int versions,
	const char *xrd_fn,
	const char *sfn,
	const char *dpmdhost,
	const char *pfn,
	const char *dpmtk,
	unsigned int flags,
	const char *dn,
	const char *vomsnfo,
	time_t tim,
	int tim_grace,
	const char *nonce,
	const XrdOucString &locstr,
	const std::vector<XrdOucString> &chunkstr,
	unsigned const char *key,
	size_t keylen);

int compareHash(
        const char *h1,
        const char *h2);

inline const char *SafeCStr(const XrdOucString &in) {
        const char *p = in.c_str();
        if (!p)
                return "";  
        return p;
}

void InitLocalHostNameList(std::vector<XrdOucString> &names);

const char *LoadKeyFromFile(unsigned char **dat, size_t *dsize);

class DpmCommonConfigOptions {
public:
	DpmCommonConfigOptions() : OssTraceLevel(0), OfsTraceLevel(0),
		DmliteConfig("/etc/dmlite.conf") { }
	int OssTraceLevel;
	int OfsTraceLevel;
	XrdOucString DmliteConfig;
};

class DpmRedirConfigOptions {
public:
	DpmRedirConfigOptions() : theN2N(0), sf(0) { }
        XrdOucString    defaultPrefix;
        std::vector<std::pair<XrdOucString, XrdOucString > >    pathPrefixes;
	DpmIdentityConfigOptions        IdentConfig;
	std::vector<XrdOucString>	AuthLibRestrict;
	XrdOucName2Name*	theN2N;
	XrdOucString	lroot_param;
	std::vector<XrdOucString>	N2NCheckPrefixes;
	XrdDmStackStore*	sf;
};

XrdOucString CanonicalisePath(const char *s, int trailing_slash);

XrdOucString TranslatePath(
	DpmRedirConfigOptions &config,
	const char *in);

void LocationToOpaque(const dmlite::Location &loc, XrdOucString &locstr,
	std::vector<XrdOucString> &chunkstr);

void EnvToLocstr(XrdOucEnv *env, XrdOucString &locstr,
	std::vector<XrdOucString> &chunkstr);

void EnvToLocation(dmlite::Location &loc, XrdOucEnv *env, const char *fn);

int DmExErrno(const dmlite::DmException &e);

XrdOucString DmExStrerror(const dmlite::DmException &e);

XrdSysError_Table *XrdDmliteError_Table();

int DpmCommonConfigProc(XrdSysError &Eroute, const char *configfn,
	DpmCommonConfigOptions &conf);

void XrdDmCommonInit();

#endif	/* XRDDPMCOMMON_HH */
